#ifndef ALPNM_H
#define ALPNM_H

#include <allegro.h>


typedef struct {

    int width;
    int height;
    int depth;
    int maxval;
    char tupletype[64];

} pam_img;


typedef enum { ALPHA_OFF = 0, ALPHA_ON = 1} is_alpha;


typedef enum { PAM_BLACKANDWHITE,
               PAM_BLACKANDWHITE_ALPHA,
               PAM_GRAYSCALE,
               PAM_GRAYSCALE_ALPHA,
               PAM_RGB,
               PAM_RGB_ALPHA } pam_mode;


typedef enum {  PBM_ASCII = 1,
                PBM_BINARY = 4,
                PGM_ASCII = 2,
                PGM_BINARY = 5,
                PPM_ASCII = 3,
                PPM_BINARY = 6 } pbm_modes;


typedef enum {  NO_ERROR,
                INVALID_CHARACTER,
                INVALID_FORMAT,
                INVALID_MODE,
                INVALID_DEPTH,
                INVALID_TOKEN,
                INVALID_TUPLTYPE,
                SIZE_EXCEEDED,
                FOPEN_FAIL_R,
                FOPEN_FAIL_W,
                CREATION_FAIL,
                MAXVAL_EXCEEDED } alpnm_error_code;


extern alpnm_error_code alpnm_error;


BITMAP *load_pnm(const char *filename);
int save_pnm(BITMAP *bmp, const char *filename, int mode, const char *comment);

BITMAP *load_pbm(const char *filename);
int save_pbm(BITMAP *bmp, const char *filename, const char *comment);
BITMAP *load_pbm_b(const char *filename);
int save_pbm_b(BITMAP *bmp, const char *filename, const char *comment);

BITMAP *load_pgm(const char *filename);
int save_pgm(BITMAP *bmp, const char *filename, const char *comment);
BITMAP *load_pgm_b(const char *filename);
int save_pgm_b(BITMAP *bmp, const char *filename, const char *comment);

BITMAP *load_ppm(const char *filename);
int save_ppm(BITMAP *bmp, const char *filename, const char *comment);
BITMAP *load_ppm_b(const char *filename);
int save_ppm_b(BITMAP *bmp, const char *filename, const char *comment);

BITMAP *load_pam(const char *filename);
int save_pam(BITMAP *bmp, const char *filename, pam_mode mode, const char *comment);
void parse_line(char *buffer, char *label, char *value);
void create_from_pbm_data(char *buffer, BITMAP *bmp, int width, int height, int depth, int maxval, int bytes, int alpha);
void create_from_pgm_data(char *buffer, BITMAP *bmp, int width, int height, int depth, int maxval, int bytes, int alpha);
void create_from_ppm_data(unsigned char *buffer, BITMAP *bmp, int width, int height, int depth, int maxval, int bytes, int alpha);

char *get_alpnm_error();


#endif
