#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <allegro.h>
#include "alpnm.h"


const int MAX_LABEL_SIZE = 8;
const int MAX_VALUE_SIZE = 64;


BITMAP *load_pam(const char *filename) {

    FILE *f;
    unsigned char line[256], *data, token[MAX_LABEL_SIZE + 1], val[MAX_VALUE_SIZE + 1];
    pam_img pamfile;
    int bytes, size;
    BITMAP *bmp;

    pamfile.width = 0;
    pamfile.height = 0;
    pamfile.depth = 0;
    pamfile.maxval = 0;
    strcpy(pamfile.tupletype, "\0");

    f = fopen(filename, "rb");
    if(!f) {

        alpnm_error = FOPEN_FAIL_R;
        return NULL;

    }

    fgets(line, sizeof(line), f);
    if (strnicmp(line, "P7", 2) != 0) {

        alpnm_error = INVALID_FORMAT;
        return NULL;

    }

    fgets(line, sizeof(line), f);
    while (strnicmp(line, "ENDHDR", 6) != 0) {

        if (line[0] != '#') {

            parse_line(line, token, val);

            if (stricmp(token, "WIDTH") == 0) {

                pamfile.width = atoi(val);
                if(pamfile.width == 0) {

                    alpnm_error = SIZE_EXCEEDED;
                    return NULL;

                }

            }
            else if (stricmp(token, "HEIGHT") == 0) {

                pamfile.height = atoi(val);
                if(pamfile.height == 0) {

                    alpnm_error = SIZE_EXCEEDED;
                    return NULL;

                }

            }
            else if (stricmp(token, "DEPTH") == 0) {

                pamfile.depth = atoi(val);
                if(pamfile.depth == 0) {

                    alpnm_error = INVALID_DEPTH;
                    return NULL;

                }

            }
            else if (stricmp(token, "MAXVAL") == 0) {

                pamfile.maxval = atoi(val);
                if(pamfile.maxval <= 0 || pamfile.maxval > 65535) {

                    alpnm_error = MAXVAL_EXCEEDED;
                    return NULL;

                }

            }
            else if (stricmp(token, "TUPLTYPE") == 0)
                strcat(pamfile.tupletype, val);
            else {

                alpnm_error = INVALID_TOKEN;
                return NULL;

            }

        } // if (line[0] != '#')

        fgets(line, sizeof(line), f);

    } // while (strnicmp(line, "ENDHDR", 6) != 0)

    if (!(stricmp(pamfile.tupletype, "BLACKANDWHITE") == 0 ||
          stricmp(pamfile.tupletype, "BLACKANDWHITE_ALPHA") == 0 ||
          stricmp(pamfile.tupletype, "GRAYSCALE") == 0 ||
          stricmp(pamfile.tupletype, "GRAYSCALE_ALPHA") == 0 ||
          stricmp(pamfile.tupletype, "RGB") == 0 ||
          stricmp(pamfile.tupletype, "RGB_ALPHA") == 0)) {

        alpnm_error = INVALID_TUPLTYPE;
        return NULL;

    }

    bmp = create_bitmap(pamfile.width, pamfile.height);
    if(!bmp) {

        alpnm_error = CREATION_FAIL;
        return NULL;

    }

    if (pamfile.maxval < 256)
        bytes = 1;
    else
        bytes = 2;

    size = pamfile.width * pamfile.height * pamfile.depth * bytes;
    data = (char *)malloc(size);
    fread(data, sizeof(char), size, f);

    if (stricmp(pamfile.tupletype, "BLACKANDWHITE") == 0)
        create_from_pbm_data(data, bmp, pamfile.width, pamfile.height, pamfile.depth, pamfile.maxval, bytes, ALPHA_OFF);
    else if (stricmp(pamfile.tupletype, "BLACKANDWHITE_ALPHA") == 0)
        create_from_pbm_data(data, bmp, pamfile.width, pamfile.height, pamfile.depth, pamfile.maxval, bytes, ALPHA_ON);
    else if (stricmp(pamfile.tupletype, "GRAYSCALE") == 0)
        create_from_pgm_data(data, bmp, pamfile.width, pamfile.height, pamfile.depth, pamfile.maxval, bytes, ALPHA_OFF);
    else if (stricmp(pamfile.tupletype, "GRAYSCALE_ALPHA") == 0)
        create_from_pgm_data(data, bmp, pamfile.width, pamfile.height, pamfile.depth, pamfile.maxval, bytes, ALPHA_ON);
    else if (stricmp(pamfile.tupletype, "RGB") == 0)
        create_from_ppm_data(data, bmp, pamfile.width, pamfile.height, pamfile.depth, pamfile.maxval, bytes, ALPHA_OFF);
    else if (stricmp(pamfile.tupletype, "RGB_ALPHA") == 0)
        create_from_ppm_data(data, bmp, pamfile.width, pamfile.height, pamfile.depth, pamfile.maxval, bytes, ALPHA_ON);

    fclose(f);
    return bmp;

}


void parse_line(char *buffer, char *label, char *value) {

    int buffer_pos = 0, label_pos = 0, value_pos = 0;

    while(isspace(buffer[buffer_pos])) buffer_pos++;

    while (!isspace(buffer[buffer_pos]) && buffer[buffer_pos] != '\0') {

        if( label_pos < MAX_LABEL_SIZE) {

            label[label_pos] = buffer[buffer_pos];
            label_pos++;

        }

        buffer_pos++;

    }

    label[label_pos] = '\0';

    while(isspace(buffer[buffer_pos])) buffer_pos++;

    while (!isspace(buffer[buffer_pos]) && buffer[buffer_pos] != '\0') {

        if( value_pos < MAX_VALUE_SIZE) {

            value[value_pos] = buffer[buffer_pos];
            value_pos++;

        }

        buffer_pos++;

    }

    value[value_pos] = '\0';

}


void create_from_pbm_data(char *buffer, BITMAP *bmp, int width, int height, int depth, int maxval, int bytes, int alpha) {

    int pos, x, y, gray, a;

    if (alpha) {

        for (y = 0; y < height; y++) {

            for (x = 0; x < width; x++) {

                pos = (width * depth * bytes * y) + (depth * bytes * x);

                gray = buffer[pos] * 255;
                a = buffer[pos + bytes];

                putpixel(bmp, x, y, makeacol(gray, gray, gray, 255 * a / maxval));

            } // for (x = 0; x < pamfile.width; x++)

        } // for (y = 0; y < pamfile.height; y++)

    }
    else {

        for (y = 0; y < height; y++) {

            for (x = 0; x < width; x++) {

                pos = (width * depth * bytes * y) + (depth * bytes * x);

                gray = buffer[pos] * 255;

                putpixel(bmp, x, y, makecol(gray, gray, gray));

            } // for (x = 0; x < pamfile.width; x++)

        } // for (y = 0; y < pamfile.height; y++)

    }

}


void create_from_pgm_data(char *buffer, BITMAP *bmp, int width, int height, int depth, int maxval, int bytes, int alpha) {

    int pos, x, y, gray, a;

    if (alpha) {

        for (y = 0; y < height; y++) {

            for (x = 0; x < width; x++) {

                pos = (width * depth * bytes * y) + (depth * bytes * x);

                gray = buffer[pos];
                a = buffer[pos + bytes];

                if (maxval > 255) {

                    gray = (gray * 256) + buffer[pos + 1];
                    a = (a * 256) + buffer[pos + bytes + 1];

                }

                putpixel(bmp, x, y, makeacol(255 * gray / maxval, 255 * gray / maxval, 255 * gray / maxval, 255 * a / maxval));

            } // for (x = 0; x < pamfile.width; x++)

        } // for (y = 0; y < pamfile.height; y++)

    }
    else {

        for (y = 0; y < height; y++) {

            for (x = 0; x < width; x++) {

                pos = (width * depth * bytes * y) + (depth * bytes * x);

                gray = buffer[pos];

                if (maxval > 255)
                    gray = (gray * 256) + buffer[pos + 1];

                putpixel(bmp, x, y, makecol(255 * gray / maxval, 255 * gray / maxval, 255 * gray / maxval));

            } // for (x = 0; x < pamfile.width; x++)

        } // for (y = 0; y < pamfile.height; y++)

    }

}


void create_from_ppm_data(unsigned char *buffer, BITMAP *bmp, int width, int height, int depth, int maxval, int bytes, int alpha) {

    int pos, x, y, r, g, b, a;

    if (alpha) {

        for (y = 0; y < height; y++) {

            for (x = 0; x < width; x++) {

                pos = (width * depth * bytes * y) + (depth * bytes * x);

                r = buffer[pos];
                g = buffer[pos + bytes];
                b = buffer[pos + (2 * bytes)];
                a = buffer[pos + (3 * bytes)];

                if (maxval > 255) {

                    r = (r * 256) + buffer[pos + 1];
                    g = (g * 256) + buffer[pos + bytes + 1];
                    b = (b * 256) + buffer[pos + (2 * bytes) + 1];
                    a = (a * 256) + buffer[pos + (3 * bytes) + 1];

                }

                putpixel(bmp, x, y, makeacol(255 * r / maxval, 255 * g / maxval, 255 * b / maxval, 255 * a / maxval));

            } // for (x = 0; x < pamfile.width; x++)

        } // for (y = 0; y < pamfile.height; y++)

    }
    else {

        for (y = 0; y < height; y++) {

            for (x = 0; x < width; x++) {

                pos = (width * depth * bytes * y) + (depth * bytes * x);

                r = buffer[pos];
                g = buffer[pos + bytes];
                b = buffer[pos + (2 * bytes)];

                if (maxval > 255) {

                    r = (r * 256) + buffer[pos + 1];
                    g = (g * 256) + buffer[pos + bytes + 1];
                    b = (b * 256) + buffer[pos + (2 * bytes) + 1];

                }

                putpixel(bmp, x, y, makecol(255 * r / maxval, 255 * g / maxval, 255 * b / maxval));

            } // for (x = 0; x < pamfile.width; x++)

        } // for (y = 0; y < pamfile.height; y++)

    }

}


int save_pam(BITMAP *bmp, const char *filename, pam_mode mode, const char *comment) {

    int x, y, depth = 1, colour, gray;
    char tupltype[9];
    FILE *f = fopen(filename, "wb");
    if(!f) {

        alpnm_error = FOPEN_FAIL_W;
        return 1;

    }

    switch (mode) {

        case PAM_BLACKANDWHITE: {

            depth = 1;
            strcpy(tupltype, "BLACKANDWHITE");
            break;
        }
        case PAM_BLACKANDWHITE_ALPHA: {

            depth = 2;
            strcpy(tupltype, "BLACKANDWHITE_ALPHA");
            break;
        }
        case PAM_GRAYSCALE: {

            depth = 1;
            strcpy(tupltype, "GRAYSCALE");
            break;
        }
        case PAM_GRAYSCALE_ALPHA: {

            depth = 2;
            strcpy(tupltype, "GRAYSCALE_ALPHA");
            break;
        }
        case PAM_RGB: {

            depth = 3;
            strcpy(tupltype, "RGB");
            break;
        }
        case PAM_RGB_ALPHA: {

            depth = 4;
            strcpy(tupltype, "RGB_ALPHA");
            break;
        }
        default: {

            alpnm_error = INVALID_MODE;
            return 1;

        }

    }

    fprintf(f, "P7\n# %s\nWIDTH %d\nHEIGHT %d\nDEPTH %d\nMAXVAL %d\nTUPLTYPE %s\nENDHDR\n",
            comment, bmp->w, bmp->h, depth, 255, tupltype);

    for (y = 0; y < bmp->h; y++) {

        for (x = 0; x < bmp->w; x++) {

            switch (mode) {

                case PAM_BLACKANDWHITE: {

                    if (getpixel(bmp, x, y) == makecol(255, 255, 255))
                        fprintf(f, "%c", 1);
                    else
                        fprintf(f, "%c", 0);
                    break;

                }
                case PAM_BLACKANDWHITE_ALPHA: {

                    colour = getpixel(bmp, x, y);
                    if (getr(colour) == 255 && getg(colour) == 255 && getb(colour) == 255)
                        fprintf(f, "%c%c", 1, geta(colour));
                    else
                        fprintf(f, "%c%c", 0, geta(colour));
                    break;

                }
                case PAM_GRAYSCALE: {

                    colour = getpixel(bmp, x, y);
                    gray = getr(colour) * 30 / 100 + getg(colour) * 59 / 100 + getb(colour) * 11 / 100;
                    fprintf(f, "%c", gray);
                    break;

                }
                case PAM_GRAYSCALE_ALPHA: {

                    colour = getpixel(bmp, x, y);
                    gray = getr(colour) * 30 / 100 + getg(colour) * 59 / 100 + getb(colour) * 11 / 100;
                    fprintf(f, "%c%c", gray, geta(colour));
                    break;

                }
                case PAM_RGB: {

                    colour = getpixel(bmp, x, y);
                    fprintf(f, "%c%c%c", getr(colour), getg(colour), getb(colour));
                    break;

                }
                case PAM_RGB_ALPHA: {

                    colour = getpixel(bmp, x, y);
                    fprintf(f, "%c%c%c%c", getr(colour), getg(colour), getb(colour), geta(colour));
                    break;

                }


            } // switch (mode)

        } // for (x = 0; x < bmp->w; x++)

    } // for (y = 0; y < bmp->h; y++)

    fclose(f);
    return 0;

}

