#include <allegro.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "alpnm.h"


BITMAP *load_pbm(const char *filename) {

    FILE *f;
    BITMAP *bmp;
    char c, format[2], c_width[5], c_height[5];
    int width, height, x, y, pos = 0;

    f = fopen(filename, "rb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_R;
        return NULL;

    }

    // read magic number
    format[0] = fgetc(f);
    format[1] = fgetc(f);

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);


        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_width[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the width
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_width[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    width = atoi(c_width);
    pos = 0;

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_height[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the height
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_height[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    height = atoi(c_height);

    bmp = create_bitmap(width, height);
    if(!bmp) {

        alpnm_error = CREATION_FAIL;
        return NULL;

    }

    for (y = 0; y < height; y++) {

        for (x = 0; x < width; x++) {

            for (;;) {

                c = fgetc(f);

                if (!isspace(c)) {  // if not whitespace

                    if (c == '1') { // make sure it's a 1 or 0

                        putpixel(bmp, x, y, makecol(0, 0, 0));
                        break;

                    }
                    else if (c == '0') { // make sure it's a 1 or 0

                        putpixel(bmp, x, y, makecol(255, 255, 255));
                        break;

                    }
                    else {

                        alpnm_error = INVALID_CHARACTER;
                        return NULL; // if any other characters, bail

                    }

                } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

            } // for (;;)

        }

    }

    fclose(f);
    return bmp;

}


int save_pbm(BITMAP *bmp, const char *filename, const char *comment) {

    FILE *f;
    int x, y, column = 0;

    f = fopen(filename, "wb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_W;
        return 1;

    }

    fprintf(f, "P1\n# %.68s\n%d %d\n", comment, bmp->w, bmp->h);

    for (y = 0; y < bmp->h; y++) {

        if (column > 68 ) {  // can't have more than 70 chars on a line.

            fprintf(f, "\n");
            column = 0;

        }

        for (x = 0; x < bmp->w; x++) {

            if (getpixel(bmp, x, y) == makecol(255, 255, 255))
                fprintf(f, "0 ");
            else
                fprintf(f, "1 ");

            column++;

        }

        fprintf(f, "\n");

    }

    fclose(f);
    return 0;

}


BITMAP *load_pbm_b(const char *filename) {

    FILE *f;
    BITMAP *bmp;
    char c, format[2], c_width[5], c_height[5];
    int width, height, x, y, pos = 0, bit;

    f = fopen(filename, "rb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_R;
        return NULL;

    }

    // read magic number
    format[0] = fgetc(f);
    format[1] = fgetc(f);

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_width[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!isspace(c))

    } // for (;;)

    // read the width
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_width[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    width = atoi(c_width);
    pos = 0;

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_height[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!isspace(c))

    } // for (;;)

    // read the height
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_height[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    height = atoi(c_height);

    bmp = create_bitmap(width, height);
    if(!bmp) {

        alpnm_error = CREATION_FAIL;
        return NULL;

    }

    for (y = 0; y < height; y++) {

        for (x = 0; x < (width / 8) + 1; x++) {

            for (;;) {

                c = fgetc(f);

                if (!isspace(c)) {  // if not whitespace

                    for (bit = 0; bit < 8; bit++) {

                        if (c & 1 << (7 - bit))
                            putpixel(bmp, x + bit, y, makecol(0, 0, 0));
                        else
                            putpixel(bmp, x + bit, y, makecol(255, 255, 255));

                    } // for (bit = 0; bit < 8; bit++)

                    break;

                } // if (!isspace(c))

            } // for (;;)

        } // for (x = 0; x < (width / 8) + 1; x++)

    } // for (y = 0; y < height; y++)

    fclose(f);
    return bmp;

}


int save_pbm_b(BITMAP *bmp, const char *filename, const char *comment) {

    FILE *f;
    int x, y, value, bit, bytes;

    f = fopen(filename, "wb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_W;
        return 1;

    }

    fprintf(f, "P4\n# %.68s\n%d %d\n", comment, bmp->w, bmp->h);

    if (bmp->w % 8 == 0)
        bytes = bmp->w / 8;
    else
        bytes = bmp->w / 8 + 1;

    for (y = 0; y < bmp->h; y++) {

        for (x = 0; x < bytes; x++) {

            value = 0;

            for (bit = 0; bit < 8; bit++) {

                if (x + bit < bmp->w) {

                    if (getpixel(bmp, (x * 8) + bit, y) != makecol(255, 255, 255))
                        value += pow(2, 7 - bit);

                }

            }

            fprintf(f, "%c", value);

        }

    }

    fclose(f);
    return 0;

}
