#include <stdio.h>
#include <allegro.h>
#include "alpnm.h"


alpnm_error_code alpnm_error = NO_ERROR;


BITMAP *load_pnm(const char *filename) {

    FILE *f;
    char format[2];

    f = fopen(filename, "rb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_R;
        return NULL;

    }

    // read magic number
    format[0] = fgetc(f);
    if (format[0] != 'P') {

        alpnm_error = INVALID_FORMAT;
        return NULL;

    }

    format[1] = fgetc(f);

    switch (format[1]) {

        case '1': return load_pbm(filename);
        case '2': return load_pgm(filename);
        case '3': return load_ppm(filename);
        case '4': return load_pbm_b(filename);
        case '5': return load_pgm_b(filename);
        case '6': return load_ppm_b(filename);
        default: {
                  alpnm_error = INVALID_FORMAT;
                  return NULL;

        }

    }

}


int save_pnm(BITMAP *bmp, const char *filename, int mode, const char *comment) {

    switch (mode) {

        case PBM_ASCII:  return save_pbm(bmp, filename, comment);
        case PBM_BINARY: return save_pbm_b(bmp, filename, comment);
        case PGM_ASCII:  return save_pgm(bmp, filename, comment);
        case PGM_BINARY: return save_pgm_b(bmp, filename, comment);
        case PPM_ASCII:  return save_ppm(bmp, filename, comment);
        case PPM_BINARY: return save_ppm_b(bmp, filename, comment);
        default: {       alpnm_error = INVALID_MODE;
                         return -1;
        }

    }

}


char *get_alpnm_error() {

    switch (alpnm_error) {

        case NO_ERROR:          return "No error.";
        case INVALID_CHARACTER: return "Invalid character read from file.";
        case INVALID_FORMAT:    return "Not a valid PNM archive.";
        case INVALID_MODE:      return "Not a valid file write mode.";
        case SIZE_EXCEEDED:     return "Maximum image size exceeded.";
        case FOPEN_FAIL_R:      return "Could not open file for reading.";
        case FOPEN_FAIL_W:      return "Could not open file for writing.";
        case CREATION_FAIL:     return "Could not create file.";
        case MAXVAL_EXCEEDED:   return "Maximum colour value exceeded.";
        default:                return "";

    }

}
