#include <allegro.h>
#include <stdio.h>
#include <ctype.h>
#include "alpnm.h"


BITMAP *load_ppm(const char *filename) {

    FILE *f;
    BITMAP *bmp;
    char c, format[2], c_width[5], c_height[5], c_maxval[5], c_red[5], c_green[5], c_blue[5];;
    int width, height, maxval, x, y, pos = 0, red, green, blue;

    f = fopen(filename, "rb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_R;
        return NULL;

    }

    // read magic number
    format[0] = fgetc(f);
    format[1] = fgetc(f);

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_width[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the width
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_width[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    width = atoi(c_width);
    pos = 0;

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_height[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the height
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_height[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    height = atoi(c_height);
    pos = 0;

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_maxval[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the maximum grey value
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_maxval[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = MAXVAL_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    maxval = atoi(c_maxval);

    bmp = create_bitmap(width, height);
    if(!bmp) {

        alpnm_error = CREATION_FAIL;
        return NULL;

    }

    for (y = 0; y < height; y++) {

        for (x = 0; x < width; x++) {

            pos = 0;

            // read red
            for (;;) {

                c = fgetc(f);

                if(feof(f)) break;

                if (!isspace(c)) {  // if not whitespace

                    if (isdigit(c)) { // make sure it's a number

                        c_red[pos] = c;
                        pos++;
                        break;

                    }
                    else {

                        alpnm_error = INVALID_CHARACTER;
                        return NULL; // if any other characters, bail

                    }

                } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

            } // for (;;)

            for (;;) {


                c = fgetc(f);

                if(feof(f)) break;

                if (isdigit(c)) { // make sure it's a number

                    c_red[pos] = c;
                    pos++;
                    if (pos > 4) {

                        alpnm_error = MAXVAL_EXCEEDED;
                        return NULL; // number too big

                    }

                }
                else if (isspace(c))  // hit whitespace, ie finished reading number
                    break;
                else {

                    alpnm_error = INVALID_CHARACTER;
                    return NULL; // if any other characters, bail

                }

            } // for (;;)

            if (pos < 4) c_red[pos] = '\0';

            pos = 0;

            // read green
            for (;;) {

                c = fgetc(f);

                if(feof(f)) break;

                if (!isspace(c)) {  // if not whitespace

                    if (isdigit(c)) { // make sure it's a number

                        c_green[pos] = c;
                        pos++;
                        break;

                    }
                else {

                    alpnm_error = INVALID_CHARACTER;
                    return NULL; // if any other characters, bail

                }
                } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

            } // for (;;)

            for (;;) {

                c = fgetc(f);

                if(feof(f)) break;

                if (isdigit(c)) { // make sure it's a number

                    c_green[pos] = c;
                    pos++;
                    if (pos > 4) {

                        alpnm_error = MAXVAL_EXCEEDED;
                        return NULL; // number too big

                    }

                }
                else if (isspace(c))  // hit whitespace, ie finished reading number
                    break;
                else {

                    alpnm_error = INVALID_CHARACTER;
                    return NULL; // if any other characters, bail

                }

            } // for (;;)

            if (pos < 4) c_green[pos] = '\0';

            pos = 0;

            // read blue
            for (;;) {

                c = fgetc(f);

                if(feof(f)) break;

                if (!isspace(c)) {  // if not whitespace

                    if (isdigit(c)) { // make sure it's a number

                        c_blue[pos] = c;
                        pos++;
                        break;

                    }
                    else {

                        alpnm_error = INVALID_CHARACTER;
                        return NULL; // if any other characters, bail

                    }

                } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

            } // for (;;)

            for (;;) {

                c = fgetc(f);

                if(feof(f)) break;

                if (isdigit(c)) { // make sure it's a number

                    c_blue[pos] = c;
                    pos++;
                    if (pos > 4) {

                        alpnm_error = MAXVAL_EXCEEDED;
                        return NULL; // number too big

                    }

                }
                else if (isspace(c))  // hit whitespace, ie finished reading number
                    break;
                else {

                    alpnm_error = INVALID_CHARACTER;
                    return NULL; // if any other characters, bail

                }

            } // for (;;)

            if (pos < 4) c_blue[pos] = '\0';

            red = atoi(c_red) * 255 / maxval;
            green = atoi(c_green) * 255 / maxval;
            blue = atoi(c_blue) * 255 / maxval;
            putpixel(bmp, x, y, makecol(red, green, blue));

        } // for (x = 0; x < width; x++) {

    } // for (y = 0; y < height; y++)

    fclose(f);
    return bmp;

}


int save_ppm(BITMAP *bmp, const char *filename, const char *comment) {

    FILE *f;
    int x, y, colour;

    f = fopen(filename, "wb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_W;
        return 1;

    }

    fprintf(f, "P3\n# %.68s\n%d %d\n255\n", comment, bmp->w, bmp->h);

    for (y = 0; y < bmp->h; y++) {

        for (x = 0; x < bmp->w; x++) {

            colour = getpixel(bmp, x, y);
            fprintf(f, "%d %d %d ", getr(colour), getg(colour), getb(colour));

        }

        fprintf(f, "\n");

    }

    fclose(f);
    return 0;

}


BITMAP *load_ppm_b(const char *filename) {

    FILE *f;
    BITMAP *bmp;
    char c, format[2], c_width[5], c_height[5], c_maxval[5];
    int width, height, maxval, x, y, pos = 0, red, green, blue;

    f = fopen(filename, "rb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_R;
        return NULL;

    }

    // read magic number
    format[0] = fgetc(f);
    format[1] = fgetc(f);

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_width[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the width
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_width[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    width = atoi(c_width);
    pos = 0;

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_height[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the height
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_height[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = SIZE_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    height = atoi(c_height);
    pos = 0;

    // ignore all whitespace and anything after a comment
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')

        if (!isspace(c)) {  // if not whitespace

            if (isdigit(c)) { // make sure it's a number

                c_maxval[pos] = c;
                pos++;
                break;

            }
            else {

                alpnm_error = INVALID_CHARACTER;
                return NULL; // if any other characters, bail

            }

        } // if (!(c == 0x09 || c == 0x0A || c == 0x0D || c == 0x20))

    } // for (;;)

    // read the maximum grey value
    for (;;) {

        c = fgetc(f);

        if (c == '#') {  // comment begins

            for (;;) {

                c = fgetc(f);
                if (c == '\n') break;  // read until new-line

            } // for (;;)

        } // if (c == '#')
        else if (isdigit(c)) { // make sure it's a number

            c_maxval[pos] = c;
            pos++;
            if (pos > 4) {

                alpnm_error = MAXVAL_EXCEEDED;
                return NULL; // number too big

            }

        }
        else if (isspace(c))  // hit whitespace, ie finished reading number
            break;
        else {

            alpnm_error = INVALID_CHARACTER;
            return NULL; // if any other characters, bail

        }

    }

    maxval = atoi(c_maxval);

    bmp = create_bitmap(width, height);
    if(!bmp) {

        alpnm_error = CREATION_FAIL;
        return NULL;

    }

    for (y = 0; y < height; y++) {

        for (x = 0; x < width; x++) {

            red = fgetc(f);

            if (maxval > 255) {

                c = fgetc(f);

                red *= 256 + c;

            }

            green = fgetc(f);

            if (maxval > 255) {

                c = fgetc(f);

                green *= 256 + c;

            }

            blue = fgetc(f);

            if (maxval > 255) {

                c = fgetc(f);

                blue *= 256 + c;

            }

            putpixel(bmp, x, y, makecol(red, green, blue));


        } // for (x = 0; x < width; x++) {

    } // for (y = 0; y < height; y++)

    fclose(f);
    return bmp;

}


int save_ppm_b(BITMAP *bmp, const char *filename, const char *comment) {

    FILE *f;
    int x, y, colour;

    f = fopen(filename, "wb");
    if (!f) {

        alpnm_error = FOPEN_FAIL_W;
        return 1;

    }

    fprintf(f, "P6\n# %.68s\n%d %d\n255\n", comment, bmp->w, bmp->h);

    for (y = 0; y < bmp->h; y++) {

        for (x = 0; x < bmp->w; x++) {

            colour = getpixel(bmp, x, y);
            fprintf(f, "%c%c%c", getr(colour), getg(colour), getb(colour));

        }

    }

    fclose(f);
    return 0;

}
